#
#
# import pytest
#
# from gpmap import GenotypePhenotypeMap
# from epistasis.models import EpistasisLinearRegression
# from ..bayesian import BayesianSampler
#
# @pytest.fixture
# def model():
#     """Create a genotype-phenotype map"""
#     wildtype = "000"
#     genotypes =  ["000", "001", "010", "100", "011", "101", "110", "111"]
#     phenotypes = [  0.0,   0.1,   0.5,   0.4,   0.2,   0.8,   0.5,   1.0]
#     stdeviations = 0.01
#     gpm = GenotypePhenotypeMap(wildtype, genotypes, phenotypes, stdeviations=stdeviations)
#     model = EpistasisLinearRegression.read_gpm(gpm, order=2).fit()
#     return model
#
# class TestBayesianSampler(object):
#
#     def test_init(self, model):
#         sampler = BayesianSampler(model)
#
#         assert hasattr(sampler, "model")
#         assert hasattr(sampler, "lnlikelihood")
#         assert hasattr(sampler, "ml_thetas")
#         assert hasattr(sampler, "lnprior")
#         assert hasattr(sampler, "ndim")
#         assert hasattr(sampler, "nwalkers")
#         assert hasattr(sampler, "sampler")
#         assert hasattr(sampler, "last_run")
#
#     def test_lnprob(self, model):
#         sampler = BayesianSampler(model)
#         output = sampler.lnprob(model.thetas, model.lnlikelihood)
#
#         assert output.dtype == float
#
#     def test_get_initial_walkers(self, model):
#         sampler = BayesianSampler(model)
#         walkers = sampler.get_initial_walkers()
#
#         assert walkers.shape == (sampler.nwalkers, sampler.ndim)
#
#     def test_sample(self, model):
#         sampler = BayesianSampler(model)
#         sampler.sample(5)
#
#         assert hasattr(sampler.sampler, "chain")
#         assert sampler.samples.shape == (5*sampler.nwalkers, sampler.ndim)
#         assert sampler.last_run != None
#         assert type(sampler.last_run) == tuple
#
#     def test_predict(self, model):
#         sampler = BayesianSampler(model)
#         sampler.sample(5)
#         p = sampler.predict()
#
#         assert p.shape == (5*sampler.nwalkers, len(model.gpm.complete_genotypes))



import numpy as np
from gpmap.gpm import GenotypePhenotypeMap
from gpmap import utils

# Local imports
from epistasis.mapping import (encoding_to_sites, assert_epistasis)
from .mapping import SimulatedEpistasisMap
from epistasis.matrix import get_model_matrix
from epistasis.utils import extract_mutations_from_genotypes
from epistasis.models.utils import XMatrixException


class BaseSimulation(GenotypePhenotypeMap):
    """ Base class for simulating genotype-phenotype maps built from epistatic
    interactions.

    Parameters
    ----------
    wildtype : str
        wildtype sequence.
    mutations : dict
        dictionary mapping each site the possible mutations
    """
    def __init__(self, wildtype, mutations, order=None,
                 model_type="global",
                 **kwargs
                 ):
        self.model_type = model_type
        self.Xbuilt = {}
        genotypes = np.array(
            utils.mutations_to_genotypes(mutations, wildtype=wildtype))
        phenotypes = np.ones(len(genotypes))

        # Initialize a genotype-phenotype map
        super(BaseSimulation, self).__init__(
            wildtype,
            genotypes,
            phenotypes,
            mutations=mutations,
            **kwargs)

        if order is not None:
            self.order = order
            self.add_epistasis()

    def add_epistasis(self):
        """Add an EpistasisMap to model.
        """
        # Build epistasis interactions as columns in X matrix.
        sites = encoding_to_sites(self.order, self.encoding_table)

        # Map those columns to epistastalis dataframe.
        self.epistasis = SimulatedEpistasisMap(gpm=self, sites=sites, values=0)

    def add_X(self, X="complete", key=None):
        """Add X to Xbuilt

        Parameters
        ----------
        X :
            see above for details.
        key : str
            name for storing the matrix.

        Returns
        -------
        X_builts : numpy.ndarray
            newly built 2d array matrix
        """
        if type(X) is str and X in ['obs', 'complete']:

            # Create a list of epistatic interaction for this model.
            if hasattr(self, "epistasis"):
                columns = self.epistasis.sites
            else:
                self.add_epistasis()
                columns = self.epistasis.sites

            # Use desired set of genotypes for rows in X matrix.
            index = self.binary

            # Build numpy array
            x = get_model_matrix(index, columns, model_type=self.model_type)

            # Set matrix with given key.
            if key is None:
                key = X

            self.Xbuilt[key] = x

        elif type(X) == np.ndarray or type(X) == pd.DataFrame:
            # Set key
            if key is None:
                raise Exception("A key must be given to store.")

            # Store Xmatrix.
            self.Xbuilt[key] = X

        else:
            raise XMatrixException("X must be one of the following: 'obs',"
                                   " 'complete', "
                                   "numpy.ndarray, or pandas.DataFrame.")

        X_built = self.Xbuilt[key]
        return X_built

    def set_coefs_order(self, order):
        """Construct a set of epistatic coefficients given the epistatic
        order."""
        # Attach an epistasis model.
        self.order = order
        self.add_epistasis()
        self.epistasis.data.values = np.zeros(self.epistasis.n)
        self.epistasis.data.values[0] = 1
        return self

    def set_coefs_sites(self, sites):
        """Construct a set of epistatic coefficients given a list of
        coefficient sites."""
        self.order = max([len(s) for s in sites])
        self.add_epistasis()
        return self

    def set_coefs(self, sites, values):
        """Set the epistatic coefs

        Parameters
        ----------
        sites : List
            List of epistatic coefficient sites.
        values : List
            list of floats representing to epistatic coefficients.
        """
        self.set_coefs_sites(sites)
        self.epistasis.data.values = values
        self.build()
        return self

    @assert_epistasis
    def set_wildtype_phenotype(self, value):
        """Set the wildtype phenotype."""
        self.epistasis.data.values[0] = value
        self.build()

    @assert_epistasis
    def set_coefs_values(self, values):
        """Set coefficient values.
        """
        self.epistasis.data.values = values
        self.build()
        return self

    @assert_epistasis
    def set_coefs_random(self, coef_range):
        """Set coefs to values drawn from a random, uniform distribution between
        coef_range.

        Parameters
        ----------
        coef_range : 2-tuple
            low and high bounds for coeff values.
        """
        # Add values to epistatic interactions
        self.epistasis.data.values = np.random.uniform(
            coef_range[0], coef_range[1], size=len(self.epistasis.sites))
        self.build()
        return self

    @assert_epistasis
    def set_coefs_decay(self):
        """Use a decaying exponential model to choose random epistatic coefficient
        values that decay/shrink with increasing order.
        """
        wt_phenotype = self.epistasis.values[0]
        for order in range(1, self.epistasis.order + 1):
            # Get epistasis map for this order.
            em = self.epistasis.get_orders(order)
            index = em.index

            # Randomly choose values for the given order
            vals = 10**(-order) * np.random.uniform(-wt_phenotype,
                                                    wt_phenotype,
                                                    size=len(index))

            # Map to epistasis object.
            self.epistasis.data.values[index[0]: index[-1] + 1] = vals
        self.build()
        return self

    @classmethod
    def from_length(cls, length, **kwargs):
        """Constructs genotype from binary sequences with given length and
        phenotypes from epistasis with a given order.

        Parameters
        ----------
        length : int
            length of the genotypes
        order : int
            order of epistasis in phenotypes.

        Returns
        -------
        GenotypePhenotypeMap
        """
        wildtype = "0" * length
        mutations = utils.genotypes_to_mutations([wildtype, "1" * length])
        return cls(wildtype, mutations, **kwargs)

    @classmethod
    def from_coefs(cls, wildtype, mutations, sites, coefs, model_type="global",
                   *args, **kwargs):
        """Construct a genotype-phenotype map from epistatic coefficients.

        Parameters
        ----------
        wildtype : str
            wildtype sequence
        mutations : dict
            dictionary mapping each site to their possible mutations.
        order : int
            order of epistasis
        coefs : list or array
            epistatic coefficients
        model_type : str
            epistatic model to use in composition matrix.
            (`'global'` or `'local'`)

        Returns
        -------
        GenotypePhenotypeMap
        """
        order = max([len(l) for l in sites])
        self = cls(wildtype, mutations,
                   model_type=model_type, *args, **kwargs)
        if len(coefs) != len(sites):
            raise Exception(
                """Number of betas does not match order/mutations given.""")
        self.set_coefs(sites, coefs)
        return self

    def build(self, values=None, **kwargs):
        """ Method for construction phenotypes from model. """
        raise Exception("""Must be implemented in subclass. """)

    def set_stdeviations(self, sigma):
        """Add standard deviations to the simulated phenotypes, which can then
        be used for sampling error in the genotype-phenotype map.

        Parameters
        ----------
        sigma : float or array-like
            Adds standard deviations to the phenotypes. If float, all
            phenotypes are given the same stdeviations. Else, array must be
            same length as phenotypes and will be assigned to each phenotype.
        """
        stdeviations = np.ones(len(self.phenotypes)) * sigma
        self.data['stdeviations'] = stdeviations
        return self



import numpy as np
from sklearn.linear_model import Ridge

from ..base import BaseModel, use_sklearn
from ..utils import arghandler

# Suppress an annoying error from scikit-learn
import warnings
warnings.filterwarnings(action="ignore", module="scipy",
                        message="^internal gelsd")


@use_sklearn(Ridge)
class EpistasisRidge(BaseModel):
    """A scikit-learn Ridge Regression class for discovering sparse
    epistatic coefficients.

    Parameters
    ----------
    order : int
        order of epistasis

    model_type : str (default="global")
        model matrix type. See publication above for more information

    alpha : float
        Constant that multiplies the L1 term. Defaults to 1.0. alpha = 0 is
        equivalent to an ordinary least square, solved by the
        EpistasisLinearRegression object.

    max_iter : int
        The maximum number of iterations.

    tol : float
        The tolerance for the optimization: if the updates are smaller than
        tol, the optimization code checks the dual gap for optimality and
        continues until it is smaller than tol.

    random_state : int
        The seed of the pseudo random number generator that selects a random
        feature to update. If int, random_state is the seed used by the random
        number generator; If RandomState instance, random_state is the random
        number generator; If None, the random number generator is the
        RandomState instance used by np.random. Used when
        selection == 'random'.

    solver : str
        See scikit learn docs for Ridge.
    """
    def __init__(
            self,
            order=1,
            model_type="global",
            alpha=1.0,
            max_iter=1000,
            tol=0.0001,
            random_state=None,
            solver='auto',
            **kwargs):
        # Set Linear Regression settings.
        self.fit_intercept = False
        self.normalize = False
        self.copy_X = True
        self.alpha = alpha
        self.max_iter = max_iter
        self.tol = tol
        self.random_state = random_state
        self.solver = solver
        self.l2_ratio = 1.0

        self.set_params(model_type=model_type, order=order)
        self.Xbuilt = {}

        # Store model specs.
        self.model_specs = dict(
            order=self.order,
            model_type=self.model_type,
            **kwargs)

    def compression_ratio(self):
        """Compute the compression ratio for the Lasso regression
        """
        vals = self.epistasis.values
        zeros = vals[vals == 0]

        numer = len(zeros)
        denom = len(vals)
        return numer/denom

    @property
    def num_of_params(self):
        n = 0
        n += self.epistasis.n
        return n

    @arghandler
    def fit(self, X=None, y=None, **kwargs):
        # If a threshold exists in the data, pre-classify genotypes
        X = np.asfortranarray(X)
        self = super(self.__class__, self).fit(X, y)

        # Link coefs to epistasis values.
        self.epistasis.values = np.reshape(self.coef_, (-1,))
        return self

    def fit_transform(self, X=None, y=None, **kwargs):
        return self.fit(X=X, y=y, **kwargs)

    @arghandler
    def predict(self, X=None):
        X = np.asfortranarray(X)
        return super(self.__class__, self).predict(X)

    @arghandler
    def predict_transform(self, X=None, y=None):
        return self.predict(X=X)

    @arghandler
    def score(self, X=None, y=None):
        X = np.asfortranarray(X)
        return super(self.__class__, self).score(X, y)

    @property
    def thetas(self):
        return self.coef_

    @arghandler
    def hypothesis(self, X=None, thetas=None):
        return np.dot(X, thetas)

    @arghandler
    def hypothesis_transform(self, X=None, y=None, thetas=None):
        return self.hypothesis(X=X, thetas=thetas)

    @arghandler
    def lnlike_of_data(
            self,
            X=None, y=None,
            yerr=None,
            thetas=None):

        # Calculate y from model.
        ymodel = self.hypothesis(X=X, thetas=thetas)

        # Return the likelihood of this model (with an L1 prior)
        return (- 0.5 * np.log(2 * np.pi * yerr**2) -
                (0.5 * ((y - ymodel)**2 / yerr**2)) -
                (self.alpha * sum(abs(thetas))))

    @arghandler
    def lnlike_transform(
            self,
            X=None,
            y=None,
            yerr=None,
            lnprior=None,
            thetas=None):
        # Update likelihood.
        lnlike = self.lnlike_of_data(X=X, y=y, yerr=yerr, thetas=thetas)
        return lnlike + lnprior



# Mapping Object for epistatic interactions int the epistasis map
#
# Author: Zach Sailer
#
# ----------------------------------------------------------
# Outside imports
# ----------------------------------------------------------

import json
import itertools as it
from functools import wraps
from collections import OrderedDict

import numpy as np
import pandas as pd

# ----------------------------------------------------------
# Local imports
# ----------------------------------------------------------

import gpmap


def assert_epistasis(method):
    """Assert that an epistasis map has been attached to the object.
    """
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if hasattr(self, "epistasis") is False:
            raise AttributeError(
                self.__name__ + " does not an epistasis attribute set yet.")
        return method(self, *args, **kwargs)
    return wrapper


def site_to_key(site, state=""):
    """Convert site to key. `state` is added to end of key."""
    if type(state) != str:
        raise Exception("`state` must be a string.")
    return ",".join([str(l) for l in site]) + state


def key_to_site(key):
    """ Convert an interaction key to label."""
    return [int(k) for k in key.split(",")]


def genotype_coeffs(genotype, order=None):
    """List the possible epistatic coefficients (as label form) for a binary
    genotype up to a given order.
    """
    if order is None:
        order = len(genotype)
    length = len(genotype)
    mutations = [i + 1 for i in range(length) if genotype[i] == "1"]
    params = [[0]]
    for o in range(1, order + 1):
        params += [list(z) for z in it.combinations(mutations, o)]
    return params

def encoding_to_sites(order, encoding_table, start_order=0):
    """Build interaction sites up to nth order given a mutation alphabet.

    Parameters
    ----------
    order : int
        order of interactions
    encoding_table  : DataFrame
        encoding table from GenotypePhenotypeMap

    Returns
    -------
    sites : list
        list of all interaction sites for system with
        sequences of a given length and epistasis with given order.
    """
    # Drop the nans (represent wildtype sites)
    t = encoding_table[["mutation_index", "genotype_index"]].dropna().astype(int)

    # Get mutations indices, grouped by their position in the genotype
    mutation_loc = t.groupby("genotype_index").groups.values()
    mutation_index = [tuple(t.loc[loc]["mutation_index"]) for loc in mutation_loc]

    # Include the intercept interaction?
    if start_order == 0:
        sites = [(0,)]
        orders = range(1, order + 1)
    else:
        sites = list()
        orders = range(start_order, order + 1)

    # Construct combinations of mutations.
    for order in orders:
        for combination in it.combinations(mutation_index, order):
            sites += list(it.product(*combination))

    return sites


class EpistasisMap(object):
    """Container object (DataFrame) for epistatic interactions.
    """
    def __init__(self,
        df=None,
        sites=None,
        values=None,
        gpm=None,
        stdeviations=None):
        if df is not None and isinstance(df, pd.DataFrame) is False:
            raise Exception("""df must be a dataframe""")

        if sites is not None and isinstance(sites, list) is False:
            raise Exception("sites must be a list of lists.")

        if df is not None:
            self._from_df(df)
        else:
            self._from_sites(sites=sites, values=values, stdeviations=stdeviations)

        self._gpm = gpm

    @property
    def gpm(self):
        if self._gpm is None:
            raise Exception("No GenotypePhenotypeMap set.")
        return self._gpm

    def _from_df(self, df):
        self.data = df

    def _from_sites(self, sites, values, stdeviations):
        data = {
            'labels': self._sites_to_keys(sites),
            'orders': self._sites_to_orders(sites),
            'sites': sites,
            'values': values,
            'stdeviations': stdeviations
        }
        self.data = pd.DataFrame(data)
        self.data.loc

    @staticmethod
    def _sites_to_keys(sites):
        return [",".join([str(c) for c in s]) for s in sites]

    @staticmethod
    def _sites_to_orders(sites):
        orders = []
        for s in sites:
            if s[0] == 0:
                orders.append(0)
            else:
                orders.append(len(s))
        return orders

    def map(self, attr1, attr2):
        """Dictionary that maps attr1 to attr2."""
        return dict(zip(getattr(self, attr1), getattr(self, attr2)))

    @classmethod
    def read_dataframe(cls, df):
        """Create an epistasis model from dataframe"""
        self = cls(df=df)
        return self

    def to_dict(self):
        """Get data as dictionary."""
        return self.data.to_dict('list')

    def to_csv(self, filename):
        """Write data to a csv file."""
        self.data.to_csv(filename)

    def to_excel(self, filename):
        """Write data to excel file."""
        self.data.to_excel(filename)

    @property
    def n(self):
        """ Return the number of Interactions. """
        return len(self.data.sites)

    @property
    def values(self):
        """ Get the values of the interaction in the system"""
        return self.data['values'].values

    @property
    def index(self):
        """ Get the interaction index in interaction matrix. """
        return self.data.index

    @property
    def sites(self):
        """ Get the interaction sites, which describe the position of
        interacting mutations in the genotypes. (type==list of lists,
        see self._build_interaction_sites)
        """
        return self.data.sites.values

    def get_label_mapper(self):
        """Builds a dictionary that maps epistatic coefs to
        the mutations they represent in a genotype-phenotype map.

        - Keys represent the epistatic coef index
        - Values are strings joining the wildtype character, genotype position,
          and mutation character.

        Returns:
        -------
        label_mapper : dict
        """
        # Make mapper dict
        t = self.gpm.encoding_table.dropna()
        label_mapper = {}
        for i, row in t.iterrows():
            label = '{}{}{}'.format(
                row.wildtype_letter,
                row.site_label,
                row.mutation_letter
            )
            label_mapper[row.mutation_index] = label
        return label_mapper

    @property
    def labels(self):
        """List of site labels. Each site is represented
        as a list of labels returned by the `label_mapper`
        (see `get_label_mapper`).
        """
        mapper = self.get_label_mapper()
        labels = ['w.t.']
        for term in self.sites[1:]:
            labels.append([mapper[site] for site in term])
        return labels

    def get_orders(self, *orders):
        """Get epistasis of a given order."""
        return EpistasisMapReference(self.data, orders)

    # ----------------------------------------------
    # Setter Functions
    # ----------------------------------------------

    def set_values(self, values, filter=None):
        if hasattr(values, "__iter__") is False:
            raise Exception("Values must be iterable.")

        if filter is None:
            self.data.values = values
        else:
            if sum(filter) != len(values):
                raise Exception("Values and filter items need to be the same length")

            self.data.loc[filter, "values"] = values

    def get(self, filter):
        return self.data.loc[filter]

    @values.setter
    def values(self, values):
        """Manually set keys. NEED TO do some quality control here. """
        self.data.values = values


class EpistasisMapReference(object):

    def __init__(self, df, orders):
        self._df = df
        self._orders= orders

    @property
    def data(self):
        return self._df.loc[(self._df.orders.isin(self._orders))]

    def map(self, attr1, attr2):
        """Dictionary that maps attr1 to attr2."""
        return dict(zip(getattr(self, attr1), getattr(self, attr2)))

    def to_dict(self):
        """Get data as dictionary."""
        return self.data.to_dict('list')

    def to_csv(self, filename):
        """Write data to a csv file."""
        self.data.to_csv(filename)

    def to_excel(self, filename):
        """Write data to excel file."""
        self.data.to_excel(filename)

    @property
    def n(self):
        """ Return the number of Interactions. """
        return len(self.data.sites)

    @property
    def values(self):
        """ Get the values of the interaction in the system"""
        return self.data['values'].values

    @property
    def index(self):
        """ Get the interaction index in interaction matrix. """
        return self.data.index

    @property
    def sites(self):
        """ Get the interaction sites, which describe the position of
        interacting mutations in the genotypes. (type==list of lists,
        see self._build_interaction_sites)
        """
        return self.data.sites.values

    def set_values(self, values):
        self._df.loc[(self._df.orders.isin(self._orders)), "values"] = values

    @values.setter
    def values(self, values):
        """Manually set keys. NEED TO do some quality control here. """
        self.set_values(values)



from functools import wraps
from ..mapping import EpistasisMap
from numpy import random

class DistributionException(Exception):
    """"""

class SimulatedEpistasisMap(EpistasisMap):
    """Just like an epistasis map, but with extra methods
    for setting epistatic coefficients
    """
    def __init__(self, gpm, df=None, sites=None, values=None, stdeviations=None):
        super().__init__(df=df, sites=sites, values=values, stdeviations=stdeviations)
        self._gpm = gpm

    @property
    def avail_distributions(self):
        return random.__all__

    def set_order_from_distribution(self, orders, dist="normal", **kwargs):
        """Sets epistatic coefficients to values drawn from a statistical distribution.

        Distributions are found in SciPy's `random` module. Kwargs are passed
        directly to these methods
        """
        # Get distribution
        try: 
            method = getattr(random, dist)
        except AttributeError:
            raise DistributionException("Distribution now found. Check the `avail_distribution` "
                                        "attribute for available distributions.")

        idx = self.data.orders.isin(orders)
        self.data.loc[idx, "values"] = method(
            size=sum(idx),
            **kwargs
        )
        self._gpm.build()

    @wraps(EpistasisMap.set_values)
    def set_values(self, values, filter=None):
        super().set_values(values, filter=filter)
        self._gpm.build()


# 
# from ..linear import *
# import numpy as np
#
#
# class TestLinearSimulation(object):
#
#     wildtype = "00"
#     mutations = {0: ["0", "1"], 1: ["0", "1"]}
#
#     def test_init(self):
#         sim = LinearSimulation(self.wildtype, self.mutations)
#         # Checks
#         assert isinstance(sim, LinearSimulation)
#
#     def test_set_coefs(self):
#         sim = LinearSimulation(self.wildtype, self.mutations)
#         sites = [[0], [1], [2]]
#         values = [0.5, 0.2, 0.4]
#         sim.set_coefs(sites, values)
#
#         # Test the correct attributes are set
#         assert hasattr(sim, "epistasis")
#         assert hasattr(sim.epistasis, "_values")
#         assert hasattr(sim.epistasis, "_sites")
#         assert sim.epistasis.values == values



"""Epistasis Pipeline module."""

import numpy as np
from ..stats import pearson
from .base import BaseModel
from .utils import arghandler

class EpistasisPipeline(list, BaseModel):
    """Construct a pipeline of epistasis models to run in series.

    This object is a subclass of a Python list. This means, all objects are
    changed in-place. This also means, you can append, prepend, remove,
    and rearrage Pipelines as you would a list.

    The models fit in order. A `fit_transform` method is called on each model
    one at a time. This method returns as transformed GenotypePhenotypeMap
    object and adds it to the next Epistasis model in the list.
    """
    @property
    def num_of_params(self):
        return sum([m.num_of_params for m in self])

    def add_gpm(self, gpm):
        self._gpm = gpm
        self[0].add_gpm(gpm)
        return self

    @property
    def gpm(self):
        return self._gpm

    @gpm.setter
    def gpm(self, gpm):
        raise Exception("Can not add gpm directly. Use `add_gpm` method.")

    def fit(self, X=None, y=None):
        """Fit pipeline.

        Parameters
        ----------
        X : array
            array of genotypes.

        y : array
            array of phentoypes.
        """
        # Fit the first model
        model = self[0]
        gpm = model.fit_transform(X=X, y=y)

        # Then fit every model afterwards.
        for model in self[1:]:
            # Get transformed genotype from modeli.
            model.add_gpm(gpm)

            # Fit model.
            try:
                gpm = model.fit_transform(X=X, y=y)
            except Exception as e:
                print("Failed with {}".format(model))
                print("Input was :")
                print("X : {}".format(X),
                print("y : {}".format(y)))
                raise e

        return self

    @arghandler
    def predict(self, X=None):
        """Predict genotypes.

        Parameters
        ----------
        X : array
            array of genotypes.
        """
        # Predict from last model in the list first.
        model = self[-1]
        ypred = model.predict_transform(X=X)

        # Then work backwards predicting/transforming until the first model.
        for model in self[-2::-1]:
            ypred = model.predict_transform(X=X, y=ypred)

        # Return predictions
        return ypred

    @arghandler
    def hypothesis(self, X=None, thetas=None):
        """Compute phenotypes of genotypes from
        given an array of model parameters.

        Parameters
        ----------
        X : array
            array of genotypes.

        thetas : array
            array of model parameters.
        """
        # Flatten thetas
        t = []
        idx = 0

        # Break up thetas into lists of lists
        for m in self:
            n = m.num_of_params
            t.append(list(thetas[idx:idx + n]))
            idx += n

        # Predict from last model in the list first.
        model = self[-1]
        thetas = t[-1]
        ypred = model.hypothesis_transform(X=X, thetas=thetas)

        # Then work backwards predicting/transforming until the first model.
        for i in range(-2, -(len(self)+1), -1):
            model = self[i]
            thetas = t[i]
            ypred = model.hypothesis_transform(X=X, y=ypred, thetas=thetas)

        # Return predictions
        return ypred

    @arghandler
    def score(self, X=None, y=None):
        """Score the model (using pearson coefficient).

        Parameters
        ----------
        X : array
            array of genotypes.

        y : array
            array of phentoypes.
        """
        ypred = self.predict(X=X)
        return pearson(y, ypred)**2

    @property
    def thetas(self):
        """Flattened array of parameters in all models (ordered by models)."""
        # All parameters in order of models.
        thetas = [m.thetas for m in self]
        return np.concatenate(thetas)

    @arghandler
    def lnlike_of_data(
            self,
            X=None,
            y=None,
            yerr=None,
            thetas=None):
        """Computes likelihood of each genotype-phenotype given model.

        Parameters
        ----------
        X

        Returns
        -------
        lnlike : array
            likelihood for each of each point.
        """
        # Flatten thetas
        t = []
        idx = 0
        # Break up thetas into lists of lists
        for m in self:
            n = m.num_of_params
            t.append(list(thetas[idx:idx + n]))
            idx += n

        # Predict from last model in the list first.
        model = self[-1]
        thetas = t[-1]
        lnlike = np.zeros(len(X))
        ypred = model.hypothesis_transform(X=X, y=y, thetas=thetas)
        lnlike = model.lnlike_transform(X=X, y=ypred, yerr=yerr, lnprior=lnlike, thetas=thetas)

        # Then work backwards predicting/transforming until the first model.
        for i in range(-2, -(len(self)+1), -1):
            model = self[i]
            thetas = t[i]
            ypred = model.hypothesis_transform(X=X, y=y, thetas=thetas)
            lnlike = model.lnlike_transform(X=X, y=ypred, yerr=yerr, lnprior=lnlike, thetas=thetas)

        # Return predictions
        return lnlike

    def lnlikelihood(
            self,
            X=None,
            y=None,
            yerr=None,
            thetas=None):
        """Compute the individal log-likelihoods for each datapoint from a set
        of model parameters.

        Parameters
        ----------
        X : list
            List of genotypes to compute likelihood.

        y : ndarray
            An array of phenotypes to transform.

        yerr : ndarray
            An array of the measured phenotypes standard deviations.

        thetas : ndarray
            array of model parameters. See thetas property for specifics.

        Returns
        -------
        lnlike : float
            log-likelihood of the model parameters.
        """
        lnlike = np.sum(self.lnlike_of_data(X=X, y=y, yerr=yerr, thetas=thetas))
        # If log-likelihood is infinite, set to negative infinity.
        if np.isinf(lnlike) or np.isnan(lnlike):
            return -np.inf
        return lnlike

    # -----------------------------------------------------------
    # Argument handlers.
    # -----------------------------------------------------------

    def _X(self, data=None, method=None):
        """Handle the X argument in this model."""
        X = data
        # If X is None, see if we saved an array.
        if X is None:
            return self.gpm.genotypes
        return X

    def _y(self, data=None, method=None):
        """Handle y arguments in this model."""
        y = data
        if y is None:
            return self.gpm.phenotypes
        return y

    def _yerr(self, data=None, method=None):
        """Handle yerr argument in this model."""
        yerr = data
        if yerr is None:
            return self.gpm.std.upper
        return yerr

    def _thetas(self, data=None, method=None):
        """Handle yerr argument in this model."""
        thetas = data
        if thetas is None:
            return self.thetas
        return thetas



import inspect
import numpy as np
from .base import BaseSimulation
from epistasis.models.nonlinear import Parameters
from epistasis.models.nonlinear.power import power_transform
from epistasis.matrix import get_model_matrix


class PowerScaleSimulation(BaseSimulation):
    """Creates a GenotypePhenotype map that exhibits a nonlinear shape created
    by a power transform function, and linear high-order epistasis that
    deviates from that scale.

    Parameters
    ----------
    wildtype:
    mutations:
    p0: list of floats
        A list of power-scale parameters. the order must be correct -->
        (lmbda, A, B).
    """

    def __init__(self, wildtype, mutations,
                 p0=[],
                 model_type='global',
                 **kwargs):

        # Set parameters
        self.model_type = model_type

        # Set the parameters -- this logic is ugly, I know. Parameters object
        # needs to be refactored and this will be cleaned up. low priority.
        self.parameters = Parameters()
        self.parameters.add(name="lmbda", value=p0[0])
        self.parameters.add(name="A", value=p0[1])
        self.parameters.add(name="B", value=p0[2])

        # Initialize base class.
        super(PowerScaleSimulation, self).__init__(wildtype, mutations,
                                                   **kwargs)

    @staticmethod
    def function(x, lmbda, A, B):
        """Power transform function."""
        return power_transform(x, lmbda, A, B)

    @classmethod
    def from_linear(cls, model, function, p0=[], **kwargs):
        """Layer nonlinear model on top of existing linear model."""
        # Initialize the class
        self = cls(model.wildtype, model.mutations, function, p0=p0, **kwargs)

        # Copy the epistasis map
        self.epistasis = model.epistasis

        # Build phenotypes with a nonlinear scale.
        self.build()
        return self

    def build(self, *args):
        """ Build nonlinear map from epistasis and function.
        """
        self.epistasis.values[0] = self.parameters['B']

        # Construct an X for the linear epistasis model
        X = self.add_X()

        # Build linear phenotypes
        self.linear_phenotypes = np.dot(X, self.epistasis.values)

        # Build nonlinear phenotypes
        self.data['phenotypes'] = self.function(
            self.linear_phenotypes, *self.parameters.values())



import json
import inspect
import numpy as np
import pandas as pd
from sklearn.preprocessing import binarize

from abc import abstractmethod, ABC, ABCMeta

# imports from gpmap dependency
from gpmap.gpm import GenotypePhenotypeMap

# Local imports
from epistasis.mapping import EpistasisMap, encoding_to_sites
from epistasis.matrix import get_model_matrix
from epistasis.utils import (extract_mutations_from_genotypes,
                             genotypes_to_X)
from .utils import XMatrixException
from sklearn.base import RegressorMixin, BaseEstimator

class SubclassException(Exception):
    """Subclass Exception for parent classes."""

def use_sklearn(sklearn_class):
    """Swap out last class in the inherited stack (Assuming its
    the BaseModel) with the AbstractModel below. Then, sandwiches
    the Sklearn class with all other base classes first, followed
    by the Sklearn class and the AbstractModel.
    """
    def mixer(cls):
        # Meta program the class
        bases = cls.__bases__[:-1]
        name = cls.__name__
        methods = dict(cls.__dict__)

        # Put Sklearn first in line of parent classes
        parents = bases + (sklearn_class, AbstractModel)

        # Rebuild class with Mixed in scikit learn.
        cls = type(name, parents, methods)
        return cls

    return mixer

class AbstractModel(ABC):
    """Abstract Base Class for all epistasis models.

    This class sets all docstrings not given in subclasses.
    """
    def __new__(self, *args, **kwargs):
        """Replace the docstrings of a subclass with docstrings in
        this base class.
        """
        # Get items in BaseModel.
        for name, member in inspect.getmembers(AbstractModel):
            # Get the docstring for this item
            doc = getattr(member, '__doc__')

            # Replace the docstring in self with basemodel docstring.
            try:
                member = getattr(self, name)
                member.__doc__ = doc

            except AttributeError:
                pass

        return super(AbstractModel, self).__new__(self)

    # --------------------------------------------------------------
    # Abstract Properties
    # --------------------------------------------------------------

    @property
    @abstractmethod
    def num_of_params(self):
        """Number of parameters in model.
        """
        raise SubclassException("Must be implemented in a subclass.")

    # --------------------------------------------------------------
    # Abstract Methods
    # --------------------------------------------------------------

    @abstractmethod
    def fit(self, X=None, y=None, **kwargs):
        """Fit model to data.

        Parameters
        ----------
        X : None, ndarray, or list of genotypes. (default=None)
            data used to construct X matrix that maps genotypes to
            model coefficients. If None, the model uses genotypes in the
            attached genotype-phenotype map. If a list of strings,
            the strings are genotypes that will be converted to an X matrix.
            If ndarray, the function assumes X is the X matrix used by the
            epistasis model.

        y : None or ndarray (default=None)
            array of phenotypes. If None, the phenotypes in the attached
            genotype-phenotype map is used.


        Returns
        -------
        self :
            The model is returned. Allows chaining methods.
        """
        raise SubclassException("Must be implemented in a subclass.")

    @abstractmethod
    def fit_transform(self, X=None, y=None, **kwargs):
        """Fit model to data and transform output according to model.

        Parameters
        ----------
        X : None, ndarray, or list of genotypes. (default=None)
            data used to construct X matrix that maps genotypes to
            model coefficients. If None, the model uses genotypes in the
            attached genotype-phenotype map. If a list of strings,
            the strings are genotypes that will be converted to an X matrix.
            If ndarray, the function assumes X is the X matrix used by the
            epistasis model.

        y : None or ndarray (default=None)
            array of phenotypes. If None, the phenotypes in the attached
            genotype-phenotype map is used.

        Returns
        -------
        gpm : GenotypePhenotypeMap
            The genotype-phenotype map object with transformed genotypes.
        """
        raise SubclassException("Must be implemented in a subclass.")

    @abstractmethod
    def predict(self, X=None):
        """Use model to predict phenotypes for a given list of genotypes.

        Parameters
        ----------
        X : None, ndarray, or list of genotypes. (default=None)
            data used to construct X matrix that maps genotypes to
            model coefficients. If None, the model uses genotypes in the
            attached genotype-phenotype map. If a list of strings,
            the strings are genotypes that will be converted to an X matrix.
            If ndarray, the function assumes X is the X matrix used by the
            epistasis model.

        Returns
        -------
        y : ndarray
            array of phenotypes.
        """
        raise SubclassException("Must be implemented in a subclass.")

    def predict_to_df(self, X=None):
        """Predict a list of genotypes and write the results to a dataframe."""
        # ------- Handle X --------------
        # Get object type.
        obj = X.__class__

        if X is None:
            X = self.gpm.genotypes

        elif obj is str and X[0] in self.gpm.mutations[0]:
            X = [X]

        elif obj is np.ndarray and X.ndim == 2:
            raise Exception("X must be a list of genotypes")             

        elif obj in [list, np.ndarray, pd.DataFrame, pd.Series]:
            pass

        else:
            raise Exception("X must be a list of genotypes.")

        # -------- Predict ---------------
        y = self.predict(X=X)

        return pd.DataFrame(dict(
            genotypes=X,
            phenotypes=y
        ))

    def predict_to_csv(self, filename, X=None):
        """Predict a list of genotypes and write the results to a CSV."""
        df = self.predict_to_df(X=X)
        df.to_csv(filename, index=False)

    def predict_to_excel(self, filename, X=None):
        """Predict a list of genotypes and write the results to a Excel."""
        df = self.predict_to_df(X=X)
        df.to_excel(filename, index=False)  

    @abstractmethod
    def predict_transform(self, X=None, y=None, **kwargs):
        """Transform a set of phenotypes according to the model.

        Parameters
        ----------
        X : None, ndarray, or list of genotypes. (default=None)
            data used to construct X matrix that maps genotypes to
            model coefficients. If None, the model uses genotypes in the
            attached genotype-phenotype map. If a list of strings,
            the strings are genotypes that will be converted to an X matrix.
            If ndarray, the function assumes X is the X matrix used by the
            epistasis model.

        y : ndarray
            An array of phenotypes to transform.

        Returns
        -------
        y_transform : ndarray
            array of phenotypes.
        """
        raise SubclassException("Must be implemented in a subclass.")

    @abstractmethod
    def hypothesis(self, X=None, thetas=None):
        """Compute phenotypes from given model parameters.

        Parameters
        ----------
        X : None, ndarray, or list of genotypes. (default=None)
            data used to construct X matrix that maps genotypes to
            model coefficients. If None, the model uses genotypes in the
            attached genotype-phenotype map. If a list of strings,
            the strings are genotypes that will be converted to an X matrix.
            If ndarray, the function assumes X is the X matrix used by the
            epistasis model.

        thetas : ndarray
            array of model parameters. See thetas property for specifics.

        Returns
        -------
        y : ndarray
            array of phenotypes predicted by model parameters.
        """
        raise SubclassException("Must be implemented in a subclass.")

    @abstractmethod
    def hypothesis_transform(self, X=None, y=None, thetas=None):
        """Transform phenotypes with given model parameters.

        Parameters
        ----------
        X : None, ndarray, or list of genotypes. (default=None)
            data used to construct X matrix that maps genotypes to
            model coefficients. If None, the model uses genotypes in the
            attached genotype-phenotype map. If a list of strings,
            the strings are genotypes that will be converted to an X matrix.
            If ndarray, the function assumes X is the X matrix used by the
            epistasis model.

        y : ndarray
            An array of phenotypes to transform.

        thetas : ndarray
            array of model parameters. See thetas property for specifics.

        Returns
        -------
        y : ndarray
            array of phenotypes predicted by model parameters.
        """
        raise SubclassException("Must be implemented in a subclass.")


    @abstractmethod
    def lnlike_of_data(
           self,
           X=None,
           y=None,
           yerr=None,
           thetas=None):
        """Compute the individUal log-likelihoods for each datapoint from a set
        of model parameters.

        Parameters
        ----------
        X : None, ndarray, or list of genotypes. (default=None)
            data used to construct X matrix that maps genotypes to
            model coefficients. If None, the model uses genotypes in the
            attached genotype-phenotype map. If a list of strings,
            the strings are genotypes that will be converted to an X matrix.
            If ndarray, the function assumes X is the X matrix used by the
            epistasis model.

        y : ndarray
            An array of phenotypes to transform.

        yerr : ndarray
            An array of the measured phenotypes standard deviations.

        thetas : ndarray
            array of model parameters. See thetas property for specifics.

        Returns
        -------
        y : ndarray
            array of phenotypes predicted by model parameters.
        """
        raise SubclassException("Must be implemented in a subclass.")

    @abstractmethod
    def lnlike_transform(
            self,
            X=None,
            y=None,
            yerr=None,
            lnprior=None,
            thetas=None):
        """Compute the individual log-likelihoods for each datapoint from a set
        of model parameters and a prior.

        Parameters
        ----------
        X : None, ndarray, or list of genotypes. (default=None)
            data used to construct X matrix that maps genotypes to
            model coefficients. If None, the model uses genotypes in the
            attached genotype-phenotype map. If a list of strings,
            the strings are genotypes that will be converted to an X matrix.
            If ndarray, the function assumes X is the X matrix used by the
            epistasis model.

        y : ndarray
            An array of phenotypes to transform.

        yerr : ndarray
            An array of the measured phenotypes standard deviations.

        lnprior : ndarray
            An array of priors for a given datapoint.

        thetas : ndarray
            array of model parameters. See thetas property for specifics.

        Returns
        -------
        y : ndarray
            array of phenotypes predicted by model parameters.
        """
        raise SubclassException("Must be implemented in a subclass.")

    def lnlikelihood(
            self,
            X=None,
            y=None,
            yerr=None,
            thetas=None):
        """Compute the individal log-likelihoods for each datapoint from a set
        of model parameters.

        Parameters
        ----------
        X : None, ndarray, or list of genotypes. (default=None)
            data used to construct X matrix that maps genotypes to
            model coefficients. If None, the model uses genotypes in the
            attached genotype-phenotype map. If a list of strings,
            the strings are genotypes that will be converted to an X matrix.
            If ndarray, the function assumes X is the X matrix used by the
            epistasis model.

        y : ndarray
            An array of phenotypes to transform.

        yerr : ndarray
            An array of the measured phenotypes standard deviations.

        thetas : ndarray
            array of model parameters. See thetas property for specifics.

        Returns
        -------
        lnlike : float
            log-likelihood of the model parameters.
        """
        lnlike = np.sum(self.lnlike_of_data(X=X, y=y, yerr=yerr, thetas=thetas))

        # If log-likelihood is infinite, set to negative infinity.
        if np.isinf(lnlike) or np.isnan(lnlike):
            return -np.inf
        return lnlike

    def add_X(self, X=None, key=None):
        """Add X to Xbuilt

        Keyword arguments for X:

        - None :
            Uses ``gpm.binary`` to construct X. If genotypes
            are missing they will not be included in fit. At the end of
            fitting, an epistasis map attribute is attached to the model
            class.


        Parameters
        ----------
        X :
            see above for details.
        key : str
            name for storing the matrix.

        Returns
        -------
        Xbuilt : numpy.ndarray
            newly built 2d array matrix
        """
        if isinstance(X, str) and X == None:

            if hasattr(self, "gpm") is False:
                raise XMatrixException("To build None, 'missing', or"
                                       "'complete' X matrix, a "
                                       "GenotypePhenotypeMap must be attached")

            # Get X columns
            columns = self.Xcolumns

            # Use desired set of genotypes for rows in X matrix.
            index = self.gpm.binary

            # Build numpy array
            x = get_model_matrix(index, columns, model_type=self.model_type)

            # Set matrix with given key.
            if key is None:
                key = X

            self.Xbuilt[key] = x

        elif type(X) == np.ndarray or type(X) == pd.DataFrame:
            # Set key
            if key is None:
                raise Exception("A key must be given to store.")

            # Store Xmatrix.
            self.Xbuilt[key] = X

        else:
            raise XMatrixException("X must be one of the following: None, "
                                   "'complete', numpy.ndarray, or "
                                   "pandas.DataFrame.")

        Xbuilt = self.Xbuilt[key]
        return Xbuilt

    def add_gpm(self, gpm):
        """Add a GenotypePhenotypeMap object to the epistasis model.
        """
        self._gpm = gpm

        # Reset Xbuilt.
        self.Xbuilt = {}

        # Construct columns for X matrix
        self.Xcolumns = encoding_to_sites(self.order, self.gpm.encoding_table)

        # Map those columns to epistastalis dataframe.
        self.epistasis = EpistasisMap(sites=self.Xcolumns, gpm=gpm)
        return self

    @property
    def gpm(self):
        """Data stored in a GenotypePhenotypeMap object."""
        return self._gpm

    # -----------------------------------------------------------
    # Argument handlers.
    # -----------------------------------------------------------

    def _X(self, data=None, method=None):
        """Handle the X argument in this model."""
        # Get object type.
        obj = data.__class__

        X = data
        # If X is None, see if we saved an array.
        if X is None:

            # Get X from genotypes
            X = genotypes_to_X(
                self.gpm.genotypes,
                self.gpm,
                order=self.order,
                model_type=self.model_type
            )

        elif obj is str and X in self.gpm.genotypes:
            single_genotype = [X]
            # Get X from genotypes
            X = genotypes_to_X(
                single_genotype,
                self.gpm,
                order=self.order,
                model_type=self.model_type
            )

        # If X is a keyword in Xbuilt, use it.
        elif obj is str and X in self.Xbuilt:
            X = self.Xbuilt[X]

        # If 2-d array, keep as so.
        elif obj is np.ndarray and X.ndim == 2:
            pass

        # If list of genotypes.
        elif obj in [list, np.ndarray, pd.DataFrame, pd.Series]:
            genotypes = X
            # Get X from genotypes
            X = genotypes_to_X(
                genotypes,
                self.gpm,
                order=self.order,
                model_type=self.model_type
            )
        else:
            raise Exception("X is invalid.")

        # Save X
        self.Xbuilt[method] = X
        return X

    def _y(self, data=None, method=None):
        """Handle y arguments in this model."""
        # Get object type.
        obj = data.__class__
        y = data

        if y is None:
            return self.gpm.phenotypes

        elif obj in [list, np.ndarray, pd.Series, pd.DataFrame]:
            return y

        else:
            raise Exception("y is invalid.")

    def _yerr(self, data=None, method=None):
        """Handle yerr argument in this model."""
        # Get object type.
        obj = data.__class__
        yerr = data
        if yerr is None:
            return self.gpm.std.upper

        elif obj in [list, np.ndarray, pd.Series, pd.DataFrame]:
            return yerr
        else:
            raise Exception("yerr is invalid.")

    def _thetas(self, data=None, method=None):
        """Handle yerr argument in this model."""
        # Get object type.
        obj = data.__class__
        thetas = data
        if thetas is None:
            return self.thetas

        elif obj in [list, np.ndarray, pd.Series, pd.DataFrame]:
            return thetas
        else:
            raise Exception("thetas is invalid.")

    def _lnprior(self, data=None, method=None):
        # Get object type.
        obj = data.__class__
        _lnprior = data
        if _lnprior is None:
            return np.zeros(self.gpm.n)

        elif obj in [list, np.ndarray, pd.Series, pd.DataFrame]:
            return _lnprior
        else:
            raise Exception("_prior is invalid.")


class BaseModel(AbstractModel, RegressorMixin, BaseEstimator):
    """Base model for defining an epistasis model.
    """
    pass



import numpy as np
import pandas as pd

from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.preprocessing import binarize

from epistasis.mapping import EpistasisMap
from epistasis.models.base import BaseModel, use_sklearn
from epistasis.models.utils import arghandler
from epistasis.models.linear import EpistasisLinearRegression

from .base import EpistasisClassifierMixin

# Use if inheriting from a scikit-learn class
@use_sklearn(GaussianProcessClassifier)
class EpistasisGaussianProcess(EpistasisClassifierMixin, BaseModel):
    """testing quadratic
    """
    def __init__(self, order=1, threshold=5, model_type='global', **kwargs):
        self.model_type = model_type
        self.order = 1
        self.Xbuilt = {}
        self.threshold=threshold

        super(self.__class__, self).__init__(**kwargs)

        # Store model specs.
        self.model_specs = dict(
            priors=None,
            threshold=threshold,
            model_type=self.model_type,
            **kwargs)

        # Set up additive linear model for pre-classifying
        self.Additive = EpistasisLinearRegression(
            order=1, model_type=self.model_type)

    @property
    def num_of_params(self):
        n = 0
        n += self.epistasis.n
        return n

    @arghandler
    def fit(self, X=None, y=None, **kwargs):
        # Use Additive model to establish the phenotypic scale.
        # Prepare Additive model
        self._fit_additive(X=X, y=y)
        self._fit_classifier(X=X, y=y)
        return self

    def hypothesis(self, X=None, thetas=None):
        pass

    def hypothesis_transform(self, X=None, y=None, thetas=None):
        pass

    def lnlike_of_data(
        self,
        X=None,
        y=None,
        yerr=None,
        thetas=None):

        pass

    def lnlike_transform(
        self,
        X=None,
        y=None,
        yerr=None,
        lnprior=None,
        thetas=None):
        pass



# from ..nonlinear import *
# import numpy as np
# from nose import tools
# 
# class TestNonlinearSimulation(object):
#     
#     wildtype = "00"
#     mutations = {0:["0","1"], 1:["0","1"]}
# 
#     def test_init(self):
#         sim = NonlinearSimulation(self.wildtype, self.mutations)
#         # Checks
#         assert isinstance(sim, NonlinearSimulation)



# Suppress an annoying error
import warnings
# warnings.filterwarnings(action="ignore", category=RuntimeWarning)

# Scipy stack imports
import numpy as np
import pandas as pd

# Scikit learn imports
from sklearn.base import BaseEstimator, RegressorMixin

# GPMap import
from gpmap import GenotypePhenotypeMap

# Epistasis imports.
from epistasis.mapping import EpistasisMap
from epistasis.models.base import BaseModel
from epistasis.models.utils import (arghandler, FittingError)
from epistasis.models.linear import (EpistasisLinearRegression, EpistasisLasso)
from epistasis.stats import pearson

from .minimizer import FunctionMinimizer

class EpistasisNonlinearRegression(BaseModel):
    """Use nonlinear least-squares regression to estimate epistatic coefficients
    and nonlinear scale in a nonlinear genotype-phenotype map.

    This models has two steps:
        1. Fit an additive, linear regression to approximate the average effect
        of individual mutations.
        2. Fit the nonlinear function to the observed phenotypes vs. the
        additive phenotypes estimated in step 1.

    Methods are described in the following publication:
        Sailer, Z. R. & Harms, M. J. 'Detecting High-Order Epistasis in
        Nonlinear Genotype-Phenotype Maps'. Genetics 205, 1079-1088 (2017).

    Parameters
    ----------
    function : callable
        Nonlinear function between Pobs and Padd

    model_type : str (default: global)
        type of epistasis model to use. See paper above for more information.

    Keyword Arguments
    -----------------
    Keyword arguments are interpreted as intial guesses for the nonlinear
    function parameters. Must have the same name as parameters in the
    nonlinear function.

    Attributes
    ----------
    Additive : EpistasisLinearRegression
        Linear regression object for fitting additive model

    parameters : Parameters object
        Mapping object for nonlinear coefficients

    minimizer :
        Object that fits data using the function and a least squares minimization.
    """
    def __init__(self,
                 function,
                 model_type="global",
                 **p0):

        # Set up the function for fitting.
        self.function = function
        self.minimizer = FunctionMinimizer(self.function, **p0)
        self.parameters = self.minimizer.parameters
        self.order = 1
        self.Xbuilt = {}

        # Construct parameters object
        self.set_params(model_type=model_type)

        # Store model specs.
        self.model_specs = dict(
            function=self.function,
            model_type=self.model_type,
            **p0)

        # Set up additive and high-order linear model
        self.Additive = EpistasisLinearRegression(
            order=1, model_type=self.model_type)

    def add_gpm(self, gpm):
        super(EpistasisNonlinearRegression, self).add_gpm(gpm)
        # Add gpm to other models.
        self.Additive.add_gpm(gpm)
        return self

    @property
    def thetas(self):
        return np.concatenate((list(self.parameters.values()),
                               self.Additive.coef_))

    @property
    def num_of_params(self):
        n = 0
        n += len(self.parameters) + len(self.Additive.coef_)
        return n

    @arghandler
    def transform(self, X=None, y=None):
        # Use a first order matrix only.
        if type(X) == np.ndarray or type(X) == pd.DataFrame:
            Xadd = X[:, :self.Additive.epistasis.n]
        else:
            Xadd = X

        # Predict additive model.
        x = self.Additive.predict(X=Xadd)

        # Transform y onto x scale
        return self.minimizer.transform(x, y)

    def fit(self,
            X=None,
            y=None,
            **kwargs):
        # Fit linear portion
        self._fit_additive(X=X, y=y)

        # Step 2: fit nonlinear function
        self._fit_nonlinear(X=X, y=y, **kwargs)
        return self

    def _fit_additive(self, X=None, y=None, **kwargs):
        # Fit with an additive model
        self.Additive.epistasis = EpistasisMap(
            sites=self.Additive.Xcolumns,
        )

        # Use a first order matrix only.
        if type(X) == np.ndarray or type(X) == pd.DataFrame:
            Xadd = X[:, :self.Additive.epistasis.n]
        else:
            Xadd = X

        # Fit Additive model
        self.Additive.fit(X=Xadd, y=y)
        self.Additive.epistasis.values = self.Additive.coef_
        return self

    @arghandler
    def _fit_nonlinear(self, X=None, y=None, **kwargs):
        """Estimate the scale of multiple mutations in a genotype-phenotype
        map."""
        # Use a first order matrix only.
        if type(X) == np.ndarray or type(X) == pd.DataFrame:
            Xadd = X[:, :self.Additive.epistasis.n]
        else:
            Xadd = X

        # Predict additive phenotypes.
        x = self.Additive.predict(X='fit')

        # Fit function
        self.minimizer.fit(x, y)
        self.parameters = self.minimizer.parameters

    @arghandler
    def fit_transform(self, X=None, y=None, **kwargs):
        self.fit(X=X, y=y, **kwargs)

        linear_phenotypes = self.transform(X=X, y=y)

        # Transform map.
        gpm = GenotypePhenotypeMap.read_dataframe(
            dataframe=self.gpm.data,
            wildtype=self.gpm.wildtype,
            mutations=self.gpm.mutations
        )

        gpm.data['phenotypes'] = linear_phenotypes
        return gpm

    def predict(self, X=None):
        x = self.Additive.predict(X=X)
        y = self.minimizer.predict(x)
        return y

    def predict_transform(self, X=None, y=None):
        if y is None:
            x = self.Additive.predict(X=X)
        else:
            x = y
        return self.minimizer.predict(x)

    @arghandler
    def hypothesis(self, X=None, thetas=None):
        # ----------------------------------------------------------------------
        # Part 0: Break up thetas
        # ----------------------------------------------------------------------
        i, j = len(self.parameters.valuesdict()), self.Additive.epistasis.n
        parameters = thetas[:i]
        epistasis = thetas[i:i + j]

        # Part 1: Linear portion
        x = np.dot(X, epistasis)

        # Part 2: Nonlinear portion
        ynonlin = self.minimizer.function(x, *parameters)

        return ynonlin

    def hypothesis_transform(self, X=None, y=None, thetas=None):
        # Break up thetas
        i, j = len(self.parameters.valuesdict()), self.Additive.epistasis.n
        parameters = thetas[:i]
        epistasis = thetas[i:i + j]

        if y is None:
            x = self.Additive.hypothesis(X=X, thetas=epistasis)
        else:
            x = y
        y_transform = self.minimizer.function(x, *parameters)
        return y_transform

    @arghandler
    def score(self, X=None, y=None):
        x = self.Additive.predict(X=X)
        ypred = self.minimizer.predict(x)
        return pearson(y, ypred)**2

    @arghandler
    def lnlike_of_data(self, X=None, y=None, yerr=None, thetas=None):
        # Calculate ymodel
        ymodel = self.hypothesis(X=X, thetas=thetas)

        # Likelihood of data given model
        return (- 0.5 * np.log(2 * np.pi * yerr**2) -
                (0.5 * ((y - ymodel)**2 / yerr**2)))

    @arghandler
    def lnlike_transform(
            self,
            X=None,
            y=None,
            yerr=None,
            lnprior=None,
            thetas=None):

        # Update likelihood.
        lnlike = self.lnlike_of_data(X=X, y=y, yerr=yerr, thetas=thetas)
        return lnlike + lnprior



import numpy as np
import pandas as pd

# Scikit-learn classifiers
from sklearn.preprocessing import binarize

from epistasis.models.base import BaseModel, use_sklearn
from epistasis.models.utils import (XMatrixException, arghandler)

from .base import EpistasisClassifierMixin

#@use_sklearn(sklearn_class)
class EpistasisClassiferClass(EpistasisClassifierMixin, BaseModel):
    """Template for EpistasisClassifier
    """
    def __init__(self, threshold, model_type="global", **kwargs):
        pass
        super().__init__(**kwargs)

        # Store model specs.
        self.model_specs = dict(
            threshold=self.threshold,
            model_type=self.model_type,
            **kwargs)

    #@arghandler
    def fit(self, X=None, y=None, **kwargs):
        # Use Additive model to establish the phenotypic scale.
        # Prepare Additive model
        self._fit_additive(X=X, y=y)
        self._fit_classifier(X=X, y=y)
        return self

    @property
    def num_of_params(self):
        pass

    #@arghandler
    def score(self, X=None, y=None, **kwargs):
        pass

    #@arghandler
    def lnlike_of_data(self, X=None, y=None, yerr=None, thetas=None):
        pass

    #@arghandler
    def lnlike_transform(
        self,
        X=None,
        y=None,
        yerr=None,
        lnprior=None,
        thetas=None):
        pass

    #@arghandler
    def hypothesis(self, X=None, thetas=None):
        pass

    def hypothesis_transform(self, X=None, y=None, thetas=None):
        pass

    @property
    def thetas(self):
        pass



import numpy as np
from sklearn.linear_model import LinearRegression

from ..base import BaseModel, use_sklearn
from ..utils import arghandler

# Suppress an annoying error from scikit-learn
import warnings
warnings.filterwarnings(action="ignore", module="scipy",
                        message="^internal gelsd")

@use_sklearn(LinearRegression)
class EpistasisLinearRegression(BaseModel):
    """Ordinary least-squares regression for estimating high-order, epistatic
    interactions in a genotype-phenotype map.

    Methods are described in the following publication:
        Sailer, Z. R. & Harms, M. J. 'Detecting High-Order Epistasis in
        Nonlinear Genotype-Phenotype Maps'. Genetics 205, 1079-1088 (2017).

    Parameters
    ----------
    order : int
        order of epistasis

    model_type : str (default="global")
        model matrix type. See publication above for more information
    """
    def __init__(self, order=1, model_type="global", n_jobs=1, **kwargs):
        # Set Linear Regression settings.
        self.fit_intercept = False
        self.normalize = False
        self.copy_X = False
        self.n_jobs = n_jobs
        self.set_params(model_type=model_type, order=order)
        self.Xbuilt = {}

        # Store model specs.
        self.model_specs = dict(
            order=self.order,
            model_type=self.model_type,
            **kwargs)

    @property
    def num_of_params(self):
        n = 0
        n += self.epistasis.n
        return n

    #@epistasis_fitter
    @arghandler
    def fit(self, X=None, y=None, **kwargs):
        self = super(self.__class__, self).fit(X, y)

        # Link coefs to epistasis values.
        self.epistasis.values = np.reshape(self.coef_, (-1,))
        return self

    def fit_transform(self, X=None, y=None, **kwargs):
        return self.fit(X=X, y=y, **kwargs)

    @arghandler
    def predict(self, X=None):
        return super(self.__class__, self).predict(X)

    def predict_transform(self, X=None, y=None):
        return self.predict(X=X)

    @arghandler
    def score(self, X=None, y=None):
        return super(self.__class__, self).score(X, y)

    @property
    def thetas(self):
        return self.coef_

    @arghandler
    def hypothesis(self, X=None, thetas=None):
        return np.dot(X, thetas)

    def hypothesis_transform(self, X=None, y=None, thetas=None):
        return self.hypothesis(X=X, thetas=thetas)

    @arghandler
    def lnlike_of_data(
            self,
            X=None,
            y=None,
            yerr=None,
            thetas=None):
        # Calculate y from model.
        ymodel = self.hypothesis(X=X, thetas=thetas)
        return (- 0.5 * np.log(2 * np.pi * yerr**2) -
                (0.5 * ((y - ymodel)**2 / yerr**2)))

    @arghandler
    def lnlike_transform(
            self,
            X=None,
            y=None,
            yerr=None,
            lnprior=None,
            thetas=None):
        # Update likelihood.
        lnlike = self.lnlike_of_data(X=X, y=y, yerr=yerr, thetas=thetas)
        return lnlike + lnprior



import pandas as pd
import numpy as np
import emcee
import warnings
from functools import wraps


class BayesianSampler(object):
    """A sampling class to estimate the uncertainties in an epistasis model's
    coefficients using a Bayesian approach. This object samples from
    the experimental uncertainty in the phenotypes to estimate confidence
    intervals for the coefficients in an epistasis model according to Bayes
    Theorem:

    .. math::
        P(H|E) = \\frac{ P(E|H) \cdot P(H) }{ P(E) }

    This reads: "the probability of epistasis model :math:`H` given the data
    :math:`E` is equal to the probability of the data given the model times the
    probability of the model."

    Parameters
    ----------
    model :
        Epistasis model to run a bootstrap calculation.
    """

    def __init__(self, model, lnprior=None):
        # Get needed features from ML model.
        self.model = model
        self.lnlikelihood = model.lnlikelihood
        self.ml_thetas = self.model.thetas

        # Set the log-prior function
        if lnprior is not None:
            self.lnprior = lnprior

        # Prepare emcee sampler
        # Get dimensions of the sampler (number of walkers, number of coefs to
        # sample)
        self.ndim = len(self.ml_thetas)
        self.nwalkers = 2 * self.ndim

        # Construct sampler
        self.sampler_engine = emcee.EnsembleSampler(
            self.nwalkers, self.ndim, self.lnprob, args=(self.lnlikelihood,))

    @staticmethod
    def lnprior(thetas):
        """Prior probability for the given set of model parameters."""
        return 0.0

    @staticmethod
    def lnprob(thetas, lnlike):
        """The posterior probability of a given set of model parameters and
        likelihood function."""
        lp = BayesianSampler.lnprior(thetas)
        if not np.isfinite(lp):
            return -np.inf
        return lp + lnlike(thetas=thetas)

    def get_initial_walkers(self, relative_widths=1e-2):
        """Place the walkers in Gaussian balls in parameter space around
        the ML values for each coefficient.
        """
        middle_positions = np.array(self.ml_thetas)
        # Construct walkers for each coefficient
        deviations = np.random.randn(self.nwalkers, self.ndim)

        # Scale deviations appropriately for coefficient magnitude.
        scales = (relative_widths * middle_positions)

        # Scale deviations
        rel_deviations = np.multiply(deviations, scales)

        # Return walker positions
        walker_positions = middle_positions + rel_deviations
        return walker_positions

    def sample(self, n_steps=100, n_burn=0, previous_state=None):
        """Sample the likelihood of the model by walking n_steps with each
        walker."""
        # Suppress warnings that occur when sampling the model.
        warnings.simplefilter("ignore", RuntimeWarning)

        # Check if a previous run was given
        if previous_state is None:
            # Get initialize positions
            pos = self.get_initial_walkers()

            # Run the MCMC walks, burning these states to equilibrate.
            if n_burn != 0:
                pos, lnprob, rstate = self.sampler_engine.run_mcmc(
                    pos0=pos, N=n_burn, storechain=False)
            else:
                lnprob, rstate = None, None
        else:
            # Get previous state.
            pos = previous_state['pos']
            lnprob = previous_state['lnprob']
            rstate = previous_state['rstate']

        # Run sampler from previous position
        pos, lnprob, rstate = self.sampler_engine.run_mcmc(pos0=pos,
                                                           N=n_steps,
                                                           rstate0=rstate,
                                                           lnprob0=lnprob,
                                                           storechain=True)

        # Store previous run in a dictionary
        previous_state = {'pos': pos, 'lnprob': lnprob, 'rstate': rstate}
        return self.sampler_engine.flatchain, previous_state



"""
A library of models to decompose high-order epistasis in genotype-phenotype
maps.
"""
# Import linear models
from .linear import (EpistasisLinearRegression,
                     EpistasisLasso,
                     EpistasisRidge,
                     EpistasisElasticNet)

# Import nonlinear models
from .nonlinear import (EpistasisNonlinearRegression,
                        EpistasisPowerTransform,
                        EpistasisSpline)

# Import classifiers
from .classifiers import (EpistasisLogisticRegression,
                          EpistasisGaussianMixture,
                          EpistasisGaussianProcess)

# Import Pipeline object fro stitching models.
from .pipeline import EpistasisPipeline



# Externel imports
import pytest

import numpy as np
from gpmap import GenotypePhenotypeMap

# Module to test
from ..ordinary import EpistasisNonlinearRegression

import warnings

# Ignore fitting warnings
warnings.simplefilter("ignore", RuntimeWarning)


@pytest.fixture
def gpm():
    """Create a genotype-phenotype map"""
    wildtype = "000"
    genotypes = ["000", "001", "010", "100", "011", "101", "110", "111"]
    phenotypes = [0.1,   0.1,   0.5,   0.4,   0.2,   0.8,   0.5,   1.0]
    stdeviations = 0.1
    return GenotypePhenotypeMap(wildtype, genotypes, phenotypes,
                                stdeviations=stdeviations)


def function(x, A, B):
    return A * x + B


class TestEpistasisNonlinearRegression(object):

    model_type = "local"

    def test_init(self, gpm):
        m = EpistasisNonlinearRegression(function=function,
                                         model_type=self.model_type)
        m.add_gpm(gpm)

        # Checks
        assert hasattr(m, 'parameters') is True
        assert hasattr(m, 'function') is True

        parameters = m.parameters

        # Checks
        assert 'A' in parameters
        assert 'B' in parameters

    def test_fit(self, gpm):

        m = EpistasisNonlinearRegression(function=function,
                                         model_type=self.model_type,
                                         A=1,
                                         B=0)
        m.add_gpm(gpm)
        m.fit()

        assert True

    def test_score(self, gpm):
        m = EpistasisNonlinearRegression(function=function,
                                         model_type=self.model_type,
                                         A=1, B=0)
        m.add_gpm(gpm)

        m.fit()
        score = m.score()
        assert 0 <= score <= 1

    def test_predict(self, gpm):
        m = EpistasisNonlinearRegression(function=function,
                                         model_type=self.model_type,
                                         A=1, B=0)
        m.add_gpm(gpm)

        m.fit()
        y = m.predict()

        # Tests
        assert True

    def test_thetas(self, gpm):
        m = EpistasisNonlinearRegression(function=function,
                                         model_type=self.model_type,
                                         A=1, B=0)
        m.add_gpm(gpm)
        m.fit()
        coefs = m.thetas
        # Tests
        assert len(coefs) == 6

    def test_hypothesis(self, gpm):

        m = EpistasisNonlinearRegression(function=function,
                                         model_type=self.model_type,
                                         A=1, B=0)
        m.add_gpm(gpm)

        m.fit()
        predictions = m.hypothesis()
        # Tests
        assert True


    def test_lnlikelihood(self, gpm):
        m = EpistasisNonlinearRegression(function=function,
                                         model_type=self.model_type,
                                         A=1, B=0)
        m.add_gpm(gpm)

        m.fit()

        # Calculate lnlikelihood
        lnlike = m.lnlikelihood()
        assert lnlike.dtype == float



import numpy as np
import pandas as pd

# Scikit-learn classifiers
from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC
from sklearn.naive_bayes import BernoulliNB
from sklearn.preprocessing import binarize

from epistasis.mapping import EpistasisMap
from epistasis.models.base import BaseModel, use_sklearn
from epistasis.models.utils import (XMatrixException, arghandler)

from epistasis.models.linear import EpistasisLinearRegression

from gpmap import GenotypePhenotypeMap

# Suppress Deprecation warning
import warnings
warnings.filterwarnings(action="ignore", module="sklearn",
                        category=DeprecationWarning)


from .base import EpistasisClassifierMixin

@use_sklearn(LogisticRegression)
class EpistasisLogisticRegression(EpistasisClassifierMixin, BaseModel):
    """Logistic regression for estimating epistatic interactions that lead to
    nonviable phenotypes. Useful for predicting viable/nonviable phenotypes.

    Parameters
    ----------
    threshold : float
        value below which phenotypes are considered nonviable.

    order : int
        order of epistasis model

    model_type : str (default="global")
        type of model matrix to use. "global" defines epistasis with respect to
        a background-averaged "genotype-phenotype". "local" defines epistasis
        with respect to the wildtype genotype.
    """
    def __init__(self, threshold, model_type="global", **kwargs):
        super(self.__class__, self).__init__(**kwargs)
        self.threshold = threshold
        self.model_type = model_type
        self.fit_intercept = False
        self.order = 1
        self.Xbuilt = {}

        # Store model specs.
        self.model_specs = dict(
            threshold=self.threshold,
            model_type=self.model_type,
            **kwargs)

    @arghandler
    def fit(self, X=None, y=None, **kwargs):
        # Use Additive model to establish the phenotypic scale.
        # Prepare Additive model
        self._fit_additive(X=X, y=y)
        self._fit_classifier(X=X, y=y)

        self.epistasis.values = self.coef_[0]
        return self

    @property
    def num_of_params(self):
        n = 0
        n += self.epistasis.n
        return n

    @arghandler
    def score(self, X=None, y=None, **kwargs):
        yclass = binarize(y.reshape(1, -1), threshold=self.threshold)[0]
        return super(self.__class__, self).score(X=X, y=yclass)

    @arghandler
    def lnlike_of_data(self, X=None, y=None, yerr=None, thetas=None):
        # Calculate Y's
        ymodel = self.hypothesis(X=X, thetas=thetas)
        ymodel_ = 1 - ymodel
        ymodel[ymodel < 0.5] = ymodel_[ymodel < 0.5]

        return np.log(ymodel)

    @arghandler
    def lnlike_transform(
        self,
        X=None,
        y=None,
        yerr=None,
        lnprior=None,
        thetas=None):
        # Update likelihood.
        ymodel = self.hypothesis(X=X, thetas=thetas)
        yclass = np.ones(len(ymodel))
        yclass[ymodel > 0.5] = 0

        lnlike = self.lnlike_of_data(X=X, y=y, yerr=yerr, thetas=thetas)
        lnprior[yclass == 0] = 0
        return lnlike + lnprior

    @arghandler
    def hypothesis(self, X=None, thetas=None):
        # Calculate probability of each class
        logit_p0 = 1 / (1 + np.exp(np.dot(X, thetas)))

        # Returns probability of class 1
        return logit_p0

    def hypothesis_transform(self, X=None, y=None, thetas=None):
        ypred = self.hypothesis(X=X, thetas=thetas)
        y[ypred > 0.5] = self.threshold
        return y

    @property
    def thetas(self):
        return self.epistasis.values
